;;======================================================================;;
;;			SERVO LIGHT DECODER				;;
;;======================================================================;;
;;									;;
;; Program:         SERVO_READ -- read servo signal	;;
;; Code:            Jindra Fucik				;;
;; Platform:        Microchip PIC16F628A, 4 Mhz				;;
;; Date:            04.03.2014						;;
;; First release:   04.03.2014						;;
;; LastDate:        04.03.2014						;;
;;									;;
;;======================================================================;;

; Minimal external components, uses internal oscilator at 4 MHz


; Servo working range 700usec - 2700usec
; with prescaller 1:16 will be 688/16=43 to 2704/16 = 169

; This program is distributed as is but WITHOUT ANY WARRANTY
; I hope you enjoy!!
;
; Revisions:
; 04.03.2014	Start of writting code


; ----- Definitions

#define		__VERNUM	D'1'
#define		__VERDAY	0x04
#define		__VERMONTH	0x03
#define		__VERYEAR	0x14


	LIST	   p=16F628A	; target processor

	;errorlevel -305
	errorlevel -302 ; warning, register in other bank

	#include p16F628a.inc


	__CONFIG  _BODEN_ON & _CP_OFF & _WDT_OFF & _MCLRE_OFF & _PWRTE_OFF & _INTRC_OSC_NOCLKOUT 


;             +---------+
; OUT2  O ---1|RA2 U RA1|18--- O OUT1
; OUT3  O ---2|RA3   RA0|17--- O OUT0
; BTN2  I ---3|RA4   RA7|16--- I
; BTN1  I ---4|RA5   RA6|15--- I
; Vss   P ---5|Vss   Vdd|14--- Vss P
; servo I ---6|INT   RB7|13--- O OUT7
; LED H O ---7|RB1   RB6|12--- O OUT6
; LED M O ---8|RB2   RB5|11--- O OUT5
; LED L O ---9|RB3   RB4|10--- O OUT4
;             +---------+
#define		SRVIN	PORTB,0			; servo signal input pin
#define		BTN1	PORTA,5			; button for set min
#define		BTN2	PORTA,4			; button for set max

#define		DEBOUNCE	COUNT,3			; button for set max

; --- Constant values

FXTAL		equ	D'4000000'		; internal oscilator

RA_TRIS         equ     0xF0			; RA4, RA5, RA6, RA7: inputs
RB_TRIS         equ     0x01			; RB0 servo input
OPTION_INI	equ   b'11000011'   ; 11000011
                                ; 1-------, PORTB pull ups - disabled
                                ; -1------, Interrupt on rising edge of RB0/INT pin
                                ; --0-----, T0CS source Fosc/4
                                ; ---0----, T0SE edge lo>hi
                                ; ----0---, PSA prescale TMR0
                                ; -----011, PS prescaler 1:16

INTC_INI	equ	0x90			; GIE, INTE enable
PIE1_INI	equ	0x00			; no PIE



; --- EEPROM Section

#define		EE_INI		0x00

E_MIN		equ	EE_INI+0x00		; Value for min
E_MAX		equ	EE_INI+0x01		; Value for max

; ----- Variables

; --- Internal RAM Section

        cblock  0x70				; all banks

INT_W		; interrupt context registers
INT_STAT	; interrupt status

EEDATA0	; data to write in EEPROM
        endc

        cblock  0x20				; bank 0

DATAVAL	; measured data
CVMIN	; value for min
CVMAX	; value for max
NORMCNT	; count of normal size packets
COUNT	; counter for debouncing
LEDSTAT ; status leds for debug
VALUE	; value for display
FLAGS	; various flags

        endc


; --- Flags

#define		NEW_DATA	FLAGS,0		; New 3 byte packet received



; --------------- Program Section --------------------------------------


		org	0x000

PowerUp:
		clrf	STATUS			; Bank 0 default
		clrf	INTCON			; Disable all interrupts
		clrf	PCLATH			; tables on page 0
		goto	INIT

; ----------------------------------------------------------------------

		org	0x004

Interrupt:
		movwf	INT_W			; save context registers
		swapf	STATUS,w
		movwf	INT_STAT
		clrf	STATUS			; interrupt uses bank 0

Int_DCC:
		btfss	INTCON,INTF		; GPIO,2 interrupt?
		goto	EndInt

		btfsc	SRVIN
		goto	Int_High_Half
Int_Low_Half:
;		movlw	0x04
;		movwf	PORTB

		movf	TMR0,w
		movwf	DATAVAL
		btfss	INTCON,T0IF	; overflow = inconsistent signal
		bsf		NEW_DATA
		bcf	INTCON,T0IF		; clear overflow flag for counting

		bcf	INTCON,INTF		
		bsf	STATUS,RP0
		bsf	OPTION_REG,INTEDG	; next interrupt on rising edge RB0
		goto	EndInt

Int_High_Half:
;		movlw	0x08
;		movwf	PORTB

		clrf	TMR0		; prepare TMR0 for new mwasure
		bcf	INTCON,T0IF		; clear overflow flag for counting

		bcf	INTCON,INTF
		bsf	STATUS,RP0
		bcf	OPTION_REG,INTEDG	; next int. on falling edge RB0
EndInt:
		swapf	INT_STAT,w		; restore context registers
		movwf	STATUS
		swapf	INT_W,f
		swapf	INT_W,w
		retfie


; ----------------------------------------------------------------------

BitPos:
		andlw	0x07
		addwf	PCL,f

		retlw	0x01
		retlw	0x02
		retlw	0x04
		retlw	0x08
		retlw	0x10
		retlw	0x20
		retlw	0x40
		retlw	0x80

; ----------------------------------------------------------------------

INIT:
		movlw	0x07
		movwf	CMCON			; set GP2:0 to digital I/O

		bsf	STATUS,RP0		; bank 1
		movlw	RA_TRIS
		movwf	TRISA
		movlw	RB_TRIS
		movwf	TRISB
		;movlw	WPU_INI			; pull-ups
		;movwf	WPU
		;clrf	IOC			; interrupt on change
		clrf	VRCON			; voltage reference off
		movlw	OPTION_INI		; Option register: no pull-up, falling GP2, no prescaler, wdt 1:1
		movwf	OPTION_REG
		movlw	PIE1_INI
		movwf	PIE1
		bcf	STATUS,RP0		; bank 0
		clrf	PIR1
;		movlw	0x01			; Timer 1 on, 1:1 , internal 
;		movlw	0x11			; Timer 1 on, 1:2 , internal 
;		movwf	T1CON

		movlw	0x20			; clear RAM
		movwf	FSR
ClearRAM:
		clrf	INDF
		incf	FSR,f
		movlw	0x60
		xorwf	FSR,w
		btfss	STATUS,Z
		goto	ClearRAM

		movlw	INTC_INI                ; Int enable, enable RA0 int, 
		movwf	INTCON					; enable RA0 external interrupt



		call	LoadCV			; load CV values
		movlw	0x02
		movwf	PORTB

; ----------------------------------------------------------------------

MainLoop:
		btfsc	NEW_DATA		; new data received?
		call	Decode			; yes, decode

		goto	MainLoop

Decode:
		bcf		NEW_DATA

		btfss	BTN1
		call	SetMin
		
		btfss	BTN2
		call	SetMax

		movf	CVMAX,w
		subwf	DATAVAL,w			; w=DATAVAL - CVMAX
		btfsc	STATUS,C		; DATAVAL is less than CVMAX
		goto	GoUP			; no - then is DATAVAL bigger or equal

		movf	DATAVAL,w
		subwf	CVMIN,w			; w=CVMIN - DATAVAL
		btfsc	STATUS,C		; CVMIN is less than DATAVAL
		goto	GoDOWN			; no - then is DATAVAL is less or equal

		btfss	DEBOUNCE
		incf	COUNT,f

		btfss	DEBOUNCE
		return

		movlw	0x04			; LED H=0 M=1 L=0
		movwf	LEDSTAT
		
OutPos
		movf	VALUE,w
		call	BitPos
		movwf	PORTA			; bits 0-3
		andlw	0xF0			; bits 4-7
		iorwf	LEDSTAT,w		; add LEDS
		movwf	PORTB
		return

GoUP
		btfss	DEBOUNCE			; if spent some time in middle position (16 pulses = 300ms)
		goto	NotYet
		incf	VALUE,f			;increment step
		movlw	0x02			; LED H=1 M=0 L=0
		movwf	LEDSTAT
		clrf	COUNT
		goto	OutPos

GoDOWN
		btfss	DEBOUNCE			; if spent some time in middle position (16 pulses = 300ms)
		goto	NotYet
		decf	VALUE,f			;increment step
		movlw	0x08			; LED H=1 M=0 L=0
		movwf	LEDSTAT
		clrf	COUNT
		goto	OutPos

NotYet
		clrf	COUNT			; clear value of count and wait for new state
		return


SetMin:
		movf	DATAVAL,w
		movwf	EEDATA0
		movwf	CVMIN
		movlw	E_MIN
		goto	SetParm
		
SetMax:
		movf	DATAVAL,w
		movwf	EEDATA0
		movwf	CVMAX
		movlw	E_MAX
		goto	SetParm
		
; ----------------------------------------------------------------------

LoadCV:
		movlw	E_MIN			; value low
		call	EE_Read
		movwf	CVMIN
		movlw	E_MAX			; value high
		call	EE_Read
		movwf	CVMAX

		return


;---------------------------------------------------------------------------

EE_Read:
		bsf	STATUS,RP0		; w=ADR
		movwf	EEADR
		bsf	EECON1,RD
		movf	EEDATA,w
		bcf	STATUS,RP0
		return

SetParm:
		call	EE_Read			; w=ADR, EEDATA0=data. Write only changes
		xorwf	EEDATA0,w
		btfsc	STATUS,Z
		return
EE_Write:		
		movf	EEDATA0,w
		bsf	STATUS,RP0
		movwf	EEDATA
		bsf	EECON1,WREN
		bcf	INTCON,GIE
		movlw	0x55
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf	EECON1,WR
		bsf	INTCON,GIE
		bcf	EECON1,WREN
EEWrite0:
		btfsc	EECON1,WR
		goto	EEWrite0
		bcf	STATUS,RP0
		return



; ----- EEPROM default values


		org	0x2100

							; on my futuba is middle 5E (94*16=1504)
		dw	0x4B			; 75*16 = 1200
		dw	0x70			; 112*16 = 1792


		org	0x2130

		dt	"servo rd"
		dt	" J.Fucik"
		dt	(__VERDAY   >> 4)  +0x30
		dt	(__VERDAY   & 0x0F)+0x30,"/"
		dt	(__VERMONTH >> 4)  +0x30
		dt	(__VERMONTH & 0x0F)+0x30,"/"
		dt	(__VERYEAR  >> 4)  +0x30
		dt	(__VERYEAR  & 0x0F)+0x30


	end